-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local indexInfo =
{
  queue       = {};
  numQueued   = 0;
  index       = 1,
  numIndexed  = 0,
  total       = 0,
  progress    = 0
};

local filteredCategories = {};
local filteredAchievements = {[0] = -1};
local selectedCategory = -1;

local blizzardAchievementsUpdate = nil;
local blizzardDisplayAchievement = nil;

local metaHooked = {}

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to add a category to the indexing queue
--
local function indexAddCategory(categoryId)
  -- Local Variables
  local achievements = NigglesDBC.achievements;
  local isQueued = false;

  -- Check if the category is currently being indexed
  if ((indexInfo.numQueued > 0) and
      (indexInfo.queue[indexInfo.numQueued] == categoryId)) then
    -- Reset the category's indexing
    indexInfo.index = 1;
  else
    -- Check if the category is already queued for indexing
    if (indexInfo.numQueued > 1) then
      for idx = 1, indexInfo.numQueued-1 do
        if (indexInfo.queue[idx] == categoryId) then
          isQueued = true;
          break;
        end
      end
    end

    -- If the category isn't queued...
    if (not isQueued) then
      -- ...add the category to the queue
      if ((indexInfo.numQueued > 0) and (indexInfo.index > 1)) then
        indexInfo.queue[indexInfo.numQueued+1] =
          indexInfo.queue[indexInfo.numQueued];
        indexInfo.queue[indexInfo.numQueued] = categoryId;
        indexInfo.numQueued = indexInfo.numQueued+1;
      else
        indexInfo.numQueued = indexInfo.numQueued+1;
        indexInfo.queue[indexInfo.numQueued] = categoryId;
      end
    end
  end

  -- Re-calculate the progress
  indexInfo.numIndexed = indexInfo.index-1;
  indexInfo.total      = 0;
  for idx = 1, indexInfo.numQueued do
    indexInfo.total = indexInfo.total+
      achievements[indexInfo.queue[idx]].count[4];
  end

  -- Update the Niggles Achievement frame, if required
  if (NigglesAchievements ~= nil) then
    -- Update the filter's status bar
    indexInfo.progress =
      math.floor(((indexInfo.numIndexed*100)/indexInfo.total)/5)*5;
    NigglesAchievements.statusBar:SetValue(indexInfo.progress);

    -- Hide the search box and show the status bar
    NigglesAchievements.searchBox:Hide();
    NigglesAchievements.statusBar:Show();

    -- Clear any filtering
    SearchBoxTemplate_OnLoad(NigglesAchievements.searchBox);
    filteredAchievements[0] = -1;
    if (AchievementFrameAchievements:IsShown()) then
      AchievementFrameAchievements_ForceUpdate();
    end
  end

  -- Set a timer to start indexing
  NigglesTimerCreate("Achievements", NigglesAchievementsIndexStep,
    nil, nil, 0.04, true);

  return;
end

--
-- Function called 'PreClick' for a meta criteria button. It clears any filter
-- before the 'OnClick' script is called to make sure that the meta criteria
-- is visible.
--
local function metaCriteriaPreClick(self, button)
  -- Clear any filtering
  filteredAchievements[0] = -1;
  NigglesAchievements.searchBox:SetText("");
  NigglesAchievements.searchBox:ClearFocus();

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to set the achievement displayed by a button in the achievements
-- list. If there are filtered achievements it will call the standard function
-- for setting the button with a modified achievement index. If the
-- achievements shouldn't be filtered, the standard function is called directly.
--
function NigglesAchievementButtonDisplay(button, category, achievement, selectionID)
  -- Local Variables
  local id = nil;

  -- Check if there are filtered achievements that should be displayed
  if ((AchievementFrame.selectedTab == 1) and
      (filteredAchievements[0] >= 0)) then
    if (achievement <= filteredAchievements[0]) then
      id = blizzardDisplayAchievement(button, category,
        filteredAchievements[achievement], selectionID);
      button.index = achievement;
    else
      button:Hide();
    end
  else
    id = blizzardDisplayAchievement(button, category, achievement, selectionID);
  end

  return id;
end

--
-- Function to process the event when an achievement is earned
--
function NigglesAchievementEarned(achievementId)
  -- Local Variables
  local catId = GetAchievementCategory(achievementId);

  -- Check the achievement's category is known
  if (NigglesDBC.achievements[catId] ~= nil) then
    -- Add the achievement's category to the queue for indexing
    indexAddCategory(catId);
  end

  return;
end

--
-- Function to process the clicking of an Achievement Alert frame
--
function NigglesAchievementsAlertOnClick(self, button, down)
  -- Local Variables
  local filterString;

  -- Check if the search functionality is enabled
  if ((NigglesDBC.settings.UiAchievementSearch == true) and
      (NigglesAchievements ~= nil)) then
    -- Check if there is any filtering
    filterString = NigglesAchievements.searchBox:GetText();
    if ((filterString ~= nil) and (filterString ~= "")) then
      -- Clear the filtering
      NigglesAchievements.searchBox:SetText("");
      filteredAchievements[0] = -1;

      -- Call the standard function again, now the achievements are unfiltered
      AchievementAlertFrame_OnClick(self, button, down);
    end
  end

  return;
end

--
-- Function to process the loading of an Achievement Alert frame
--
function NigglesAchievementsAlertOnLoad(self)
  -- Hook the 'OnClick' event
  self:HookScript("OnClick", NigglesAchievementsAlertOnClick);

  return;
end

--
-- Function to update the Categories by greying out those categories with no
-- filter matches
--
function NigglesAchievementsCategoryUpdate()
  -- Local Variables
  local buttons = AchievementFrameCategoriesContainer.buttons;
  local button;

  -- Check if the categories should be updated
  if ((NigglesDBC.settings.UiAchievementSearch == true) and
      (buttons ~= nil)) then
    -- Change the font of the buttons for categories with no matches
    for idx = 1, #(buttons) do
      -- Check if the button is visible
      button = buttons[idx];
      if (button:IsShown()) then
        -- Check if the button's category has no matches
        if (filteredCategories[button.categoryID] == false) then
          button.label:SetFontObject("GameFontDisable");
        else
          if (type(button.element.parent) == "number") then
            button.label:SetFontObject("GameFontHighlight");
          else
            button.label:SetFontObject("GameFontNormal");
          end
        end
      end
    end
  end

  return;
end

--
-- Function to enable/disable the Niggles Achievement functionality
--
function NigglesAchievementsEnable(resetIndexing)
  -- Local Variables
  local achievements = NigglesDBC.achievements;
  local count = {0, 0, 0, 0, 0, 0};
  local edge;
  local frame;
  local isChanged;
  local texture;
  local idx;

  -- Check if any existing indexing should be reset
  if (resetIndexing == true) then
    -- Stop any existing indexing
    NigglesTimerDestroy("Achievements");

    -- Clear the index info
    indexInfo.numQueued   = 0;
    indexInfo.index       = 1;
    indexInfo.numIndexed  = 0;
    indexInfo.total       = 0;
    indexInfo.progress    = 0;

    -- Check if the achievements should be indexed
    if (NigglesDBC.settings.UiAchievementSearch == true) then
      -- Check if any of the categories need to be indexed
      for _, catId in ipairs(GetCategoryList()) do
        -- Make sure the category's info exists
        if (achievements[catId] == nil) then
          achievements[catId] = {};
        end
        if (achievements[catId].count == nil) then
          achievements[catId].count = {0, 0, 0, 0, 0, 0};
        end
        if (achievements[catId].text == nil) then
          achievements[catId].text = {};
        end
        if (achievements[catId].isIndexed == nil) then
          achievements[catId].isIndexed = false;
        end
        _, achievements[catId].parentId = GetCategoryInfo(catId);

        -- Get the number of total and completed achievements
        count[1], count[2], count[3] = GetCategoryNumAchievements(catId, true);
        count[4], count[5], count[6] = GetCategoryNumAchievements(catId, false);

        -- Search for any changes in the number of achievements
        isChanged = false;
        for idx, value in ipairs(count) do
          if (achievements[catId].count[idx] ~= value) then
            achievements[catId].count[idx] = value;
            isChanged = true;
          end
        end

        -- Queue the category for indexing, if required
        if ((isChanged) or (not achievements[catId].isIndexed)) then
          indexAddCategory(catId);
        end
      end
    end
  end

  -- Check if the Niggles Achievement frame should be created
  if ((AchievementFrame ~= nil) and
      (NigglesDBC.settings.UiAchievementSearch == true) and
      (NigglesAchievements == nil)) then
    -- Create the Niggles Achievement frame
    frame = CreateFrame("FRAME", "NigglesAchievements",
      AchievementFrame, "NigglesAchievementsTemplate");
    frame:SetBackdropBorderColor(
      ACHIEVEMENTUI_GOLDBORDER_R, ACHIEVEMENTUI_GOLDBORDER_G,
      ACHIEVEMENTUI_GOLDBORDER_B, ACHIEVEMENTUI_GOLDBORDER_A);
    frame.statusBar:SetMinMaxValues(0, 100);
    frame.statusBar:SetValue(0);
    frame.statusBar:SetStatusBarColor(0, .8, 0, 1);
    frame.statusBar.progress:SetText("0%");
    texture = frame.statusBar:GetStatusBarTexture();
    texture:SetDrawLayer("BORDER");

    -- Create a new edge to replace the existing backdrop
    edge = CreateFrame("FRAME", "NigglesAchievementsEdge",
      AchievementFrameCategories);
    edge:SetPoint("TOPLEFT", 0, -37);
    edge:SetPoint("BOTTOMRIGHT", 0, 0);
    edge:SetBackdrop(AchievementFrameCategories:GetBackdrop());
    edge:SetBackdropBorderColor(ACHIEVEMENTUI_GOLDBORDER_R,
      ACHIEVEMENTUI_GOLDBORDER_G, ACHIEVEMENTUI_GOLDBORDER_B,
      ACHIEVEMENTUI_GOLDBORDER_A);
    AchievementFrameCategories:SetBackdrop(nil);
    AchievementFrameCategoriesContainer:SetPoint("TOPLEFT", 0, -42);

    -- Hook the function that is called when the tabs change
    hooksecurefunc("AchievementFrame_UpdateTabs",
      NigglesAchievementsTabsUpdated);

    -- Hook the function that is called when the filter is changed
    hooksecurefunc("AchievementFrame_SetFilter",
      NigglesAchievementsFilterCategories);

    -- Hook the showing of the comparison frame so the search can be disabled
    AchievementFrameComparison:HookScript("OnShow",
      NigglesAchievementsTabsUpdated)

    -- Hook the function that updates the categories
    hooksecurefunc("AchievementFrameCategories_Update",
      NigglesAchievementsCategoryUpdate);
    AchievementFrameCategoriesContainer.update =
      AchievementFrameCategories_Update;

    -- Hook the function that creates the buttons to display meta achievements
    hooksecurefunc("AchievementButton_GetMeta",
      NigglesAchievementsMetaBtnCreated);

    -- Save the Blizzard functions used to update displaying of achievements
    blizzardDisplayAchievement = AchievementButton_DisplayAchievement;
    blizzardAchievementsUpdate = AchievementFrameAchievements_Update;

    -- Show the indexing or search box
    if (next(indexInfo.queue) ~= nil) then
      NigglesAchievements.searchBox:Hide();
      NigglesAchievements.statusBar:SetValue(indexInfo.progress);
      NigglesAchievements.statusBar:Show();
    else
      NigglesAchievements.statusBar:Hide();
      NigglesAchievements.searchBox:Show();
    end
  end

  -- Check the Niggles Achievement frame exists
  if (NigglesAchievements ~= nil) then
    -- Replace/restore Blizzard functions for update the Achievement frame
    if (NigglesDBC.settings.UiAchievementSearch == true) then
      -- Replace the Blizzard functions used to display achievement
      AchievementButton_DisplayAchievement = NigglesAchievementButtonDisplay;
      AchievementFrameAchievements_Update  = NigglesAchievementsUpdate;
    else
      -- Restore the Blizzard functions used to display achievement
      AchievementButton_DisplayAchievement = blizzardDisplayAchievement;
      AchievementFrameAchievements_Update  = blizzardAchievementsUpdate;
    end

    -- Update the function pointers in the Blizzard code
    AchievementFrameAchievementsContainer.update =
      AchievementFrameAchievements_Update;
    ACHIEVEMENT_FUNCTIONS.updateFunc       =
      AchievementFrameAchievements_Update;
    GUILD_ACHIEVEMENT_FUNCTIONS.updateFunc =
      AchievementFrameAchievements_Update;
    AchievementFrame_SetFilter(AchievementFrameFilterDropDown.value);

    -- Clear any filtering
    SearchBoxTemplate_OnLoad(NigglesAchievements.searchBox);
    filteredAchievements[0] = -1;
    if (AchievementFrameAchievements:IsShown()) then
      AchievementFrameAchievements_ForceUpdate();
    end

    -- Show/hide the Niggles Achievement frame
    NigglesAchievementsTabsUpdated();
  end

  return;
end

--
-- Function to filter the achievements in a category
--
function NigglesAchievementsFilterAchievements(categoryId)
  -- Local Variables
  local achievementText = (NigglesDBC.achievements[categoryId].text or {});
  local filterIdx = 0;
  local filterString;

  -- Save the ID of the selected category
  selectedCategory = categoryId;

  -- Get the filter string
  filterString = NigglesAchievements.searchBox:GetText();
  if ((filterString == SEARCH) or (filterString == "")) then
    filterString = nil;
  else
    filterString = filterString:lower();
  end

  -- Check if there is a filter string
  if (filterString ~= nil) then
    -- Search for achievements in the category that match the filter
    if (AchievementFrameFilterDropDown.value ==
         ACHIEVEMENT_FILTER_ALL) then
      for idx = 1, #achievementText do
        if (achievementText[idx]:find(filterString, 3, true) ~= nil) then
          filterIdx = filterIdx+1;
          filteredAchievements[filterIdx] = idx;
        end
      end
    elseif (AchievementFrameFilterDropDown.value ==
         ACHIEVEMENT_FILTER_COMPLETE) then
      for idx = 1, #achievementText do
        if ((achievementText[idx]:sub(1,1) == "c") and
            (achievementText[idx]:find(filterString, 3, true) ~= nil)) then
          filterIdx = filterIdx+1;
          filteredAchievements[filterIdx] = idx;
        end
      end
    elseif (AchievementFrameFilterDropDown.value ==
         ACHIEVEMENT_FILTER_INCOMPLETE) then
      for idx = 1, #achievementText do
        if ((achievementText[idx]:sub(2,2) == "n") and
            (achievementText[idx]:find(filterString, 3, true) ~= nil)) then
          filterIdx = filterIdx+1;
          filteredAchievements[filterIdx] = idx;
        end
      end
    end

    -- Save the number of achievements that match the filter
    filteredAchievements[0] = filterIdx;
  else
    -- Set the number of achievements to -1 to indicate all match
    filteredAchievements[0] = -1;
  end

  return;
end

--
-- Function to filter the achievement categories
--
function NigglesAchievementsFilterCategories()
  -- Local Variables
  local achievements = NigglesDBC.achievements;
  local achievementText;
  local ancestorId;
  local filterString;
  local filterFlag;
  local initialValue;

  -- Check if the categories should be filtered
  if ((NigglesDBC.settings.UiAchievementSearch == true) and
      (NigglesAchievements ~= nil)) then
    -- Get the filter string
    filterString = NigglesAchievements.searchBox:GetText();
    if ((filterString == SEARCH) or (filterString == "")) then
      filterString = nil;
      initialValue = true;
    else
      filterString = filterString:lower();
      initialValue = false;
    end

    -- Initialise the flags that indicates which category match the filter
    filteredCategories["summary"] = initialValue;
    for catId, _ in pairs(achievements) do
      filteredCategories[catId] = initialValue;
    end

    -- Make sure there is a filter string
    if (filterString ~= nil) then
      -- Search all categories for those that match the filter
      for catId, catInfo in pairs(achievements) do
        -- Check category hasn't already matched the filter & has achievements
        if ((filteredCategories[catId] == false) and
            (achievements[catId] ~= nil)) then
          -- Search for any achievement in the category that match the filters
          achievementText = (achievements[catId].text or {});
          if (AchievementFrameFilterDropDown.value ==
               ACHIEVEMENT_FILTER_ALL) then
            for idx = 1, #achievementText do
              if (achievementText[idx]:find(filterString, 3, true) ~= nil) then
                -- Flag the category and all its ancestors as matches
                ancestorId = catId;
                while (ancestorId ~= -1) do
                  filteredCategories[ancestorId] = true;
                  ancestorId = achievements[ancestorId].parentId;
                end
                break;
              end
            end
          elseif (AchievementFrameFilterDropDown.value ==
                    ACHIEVEMENT_FILTER_COMPLETE) then
            for idx = 1, #achievementText do
              if ((achievementText[idx]:sub(1,1) == "c") and
                  (achievementText[idx]:find(filterString, 3, true) ~=
                    nil)) then
                -- Flag the category and all its ancestors as matches
                ancestorId = catId;
                while (ancestorId ~= -1) do
                  filteredCategories[ancestorId] = true;
                  ancestorId = achievements[ancestorId].parentId;
                end
                break;
              end
            end
          elseif (AchievementFrameFilterDropDown.value ==
                    ACHIEVEMENT_FILTER_INCOMPLETE) then
            for idx = 1, #achievementText do
              if ((achievementText[idx]:sub(2,2) == "n") and
                  (achievementText[idx]:find(filterString, 3, true) ~=
                    nil)) then
                -- Flag the category and all its ancestors as matches
                ancestorId = catId;
                while (ancestorId ~= -1) do
                  filteredCategories[ancestorId] = true;
                  ancestorId = achievements[ancestorId].parentId;
                end
                break;
              end
            end
          end
        end
      end
    end

    -- Update the categories
    -- NOTE: Normally this would be done by calling the Blizzard function. On
    --       this occasion, the Niggles function is called instead. This is
    --       because the Blizzard function is hideously inefficient.
    NigglesAchievementsCategoryUpdate();

    -- Update the achievements
    if ((selectedCategory ~= -1) and
        (achievements[selectedCategory] ~= nil)) then
      NigglesAchievementsFilterAchievements(selectedCategory);
    end
    AchievementFrameAchievements_ForceUpdate();
  end

  return;
end

--
-- Function to perform a step in indexing the queued categories
--
function NigglesAchievementsIndexStep(arg1, arg2)
  -- Local Variables
  local catId = 0;
  local catInfo;
  local id;
  local description;
  local name;
  local reward;
  local numCriteria;
  local string;
  local flags;
  local completed;
  local progress;

  -- Get the category ID of the next achievement to be indexed
  while ((indexInfo.numQueued > 0) and (catId == 0)) do
    catInfo = NigglesDBC.achievements[indexInfo.queue[indexInfo.numQueued]];
    if (indexInfo.index <= catInfo.count[4]) then
      catId = indexInfo.queue[indexInfo.numQueued];
    else
      -- Remove any superfluous strings from the category's achievement table
      for idx = #(catInfo.text), catInfo.count[4]+1, -1 do
        catInfo.text[idx] = nil;
      end

      -- Remove the category from the queue
      indexInfo.numQueued = indexInfo.numQueued-1;

      -- Reset the achievement index for the category
      indexInfo.index = 1;

      -- Set the flag that indicates the category has been indexed
      catInfo.isIndexed = true;
    end
  end

  -- Check a category was identified
  if (catId ~= 0) then
    -- Check if this is the first achievement indexed in the category
    if (indexInfo.index == 1) then
      -- Reset the flag that indicates the category is indexed
      catInfo.isIndexed = false;
    end

    -- Get the achievement's info
    id, name, _, completed, _, _, _, description, flags, _, reward, _,
      earnedByMe = GetAchievementInfo(catId, indexInfo.index);

    -- Check the achievement is displayed
    if ((id ~= nil) and
        (flags ~= nil) and
        (bit.band(flags, 0x00000002) == 0)) then
      -- Build a string from the achievement's text
      string = (completed and "c" or "i")..(earnedByMe and "y" or "n")..
        (name or "").."|"..(description or "").."|"..(reward or "");

      -- Append any text for the achievement's criteria
      numCriteria = GetAchievementNumCriteria(id);
      for criteriaIdx = 1, numCriteria do
        -- Get the criterion's info
        subString, _, _, _, _, _, flags =
          GetAchievementCriteriaInfo(id, criteriaIdx);

        -- Check the criterion should be displayed
        if ((subString ~= nil) and (bit.band(flags, 0x00000002) == 0)) then
          string = string.."|"..subString;
        end
      end

      -- Save the achievement's text
      catInfo.text[indexInfo.index] = string:lower();
    end

    -- Update the index of the next achievement
    indexInfo.index      = indexInfo.index+1;
    indexInfo.numIndexed = indexInfo.numIndexed+1;

    -- Update the status bar, if required
    progress = math.floor(((indexInfo.numIndexed*100)/indexInfo.total)/5)*5
    if (progress ~= indexInfo.progress) then
      indexInfo.progress = progress;
      if (NigglesAchievements ~= nil) then
        NigglesAchievements.statusBar:SetValue(progress);
      end
    end
  end

  -- Check if all queued categories have been indexed
  if (indexInfo.numQueued == 0) then
    -- Stop the indexing
    NigglesTimerDestroy("Achievements");

    -- Hide the status bar and show the search box
    if (NigglesAchievements ~= nil) then
      NigglesAchievements.statusBar:Hide();
      NigglesAchievements.searchBox:Show();
    end
  end

  return;
end

--
-- Function to hook the 'PreClick' script of new meta criteria buttons
--
function NigglesAchievementsMetaBtnCreated(index)
  -- Check if the meta criteria button is already hooked
  if (not metaHooked[index]) then
    -- Hook the button's 'PreClick' script
    _G["AchievementFrameMeta"..index]:HookScript("PreClick",
      metaCriteriaPreClick);
    metaHooked[index] = true;
  end

  return;
end

--
-- Function called when the Niggles Achievement frame is shown
--
function NigglesAchievementsOnShow(self)
  NigglesAchievements.searchBox:SetShown(indexInfo.numQueued <= 0);
  NigglesAchievements.statusBar:SetShown(indexInfo.numQueued > 0);

  return;
end

--
-- Function to shut down the Niggles Achievements functionality
--
function NigglesAchievementsShutdown()
  -- Check if the indexed achievements should be destroyed
  if ((NigglesDBC.settings.UiAchievementSearch == false) and
      (next(NigglesDBC.achievements))) then
    wipe(NigglesDBC.achievements);
  end

  return
end
--
-- Function called when the tabs are updated
--
function NigglesAchievementsTabsUpdated()
  -- Check if the first tab is being displayed without the comparison frame
  if ((NigglesDBC.settings.UiAchievementSearch == true) and
      ((AchievementFrame.selectedTab == 1) or
       (AchievementFrame.selectedTab == nil)) and
      (AchievementFrame.isComparison ~= true)) then
    NigglesAchievementsFilterCategories();
    NigglesAchievements:Show();
    NigglesAchievementsEdge:SetPoint("TOPLEFT", 0, -37);
    AchievementFrameCategoriesContainer:SetPoint("TOPLEFT", 0, -42);
  else
    NigglesAchievements:Hide();
    NigglesAchievementsEdge:SetPoint("TOPLEFT", 0, 0);
    AchievementFrameCategoriesContainer:SetPoint("TOPLEFT", 0, -5);
    filteredAchievements[0] = -1;
  end

  return;
end

--
-- Function to update the Achievements
--
-- NOTE: This is a modified version of 'AchievementFrameAchievements_Update()'
--       in 'Blizzard_AchievementUI.lua'.
--
function NigglesAchievementsUpdate()
  -- Local Variables
  local catId = ACHIEVEMENT_FUNCTIONS.selectedCategory;
  local scrollFrame = AchievementFrameAchievementsContainer
  local offset = HybridScrollFrame_GetOffset(scrollFrame);
  local numButtons = #(scrollFrame.buttons);
  local numAchievements;
  local selection = AchievementFrameAchievements.selection;
  local extraHeight;
  local index;
  local displayedHeight = 0;
  local totalHeight;

  -- Check if indexing or any tab but the 'Achievements' tab is selected
  if ((indexInfo.numQueued > 0) or
      (AchievementFrame.selectedTab ~= 1)) then
    blizzardAchievementsUpdate();
    return;
  end

  -- Check if the 'Summary' is being displayed
  if (catId == "summary") then
    return;
  end

  -- Filter the achievements, if required
  -- NOTE: The category is not set by one function in the Blizzard code so
  --       can't be hooked, hence the call here to filter achievements
  if (catId ~= selectedCategory) then
    NigglesAchievementsFilterAchievements(catId);
  end

  -- Check if unfiltered results should be displayed
  if (filteredAchievements[0] == -1) then
    -- Call the standard Blizzard function
    blizzardAchievementsUpdate();
    return;
  end

  -- Check if there are no matches
  if (filteredAchievements[0] == 0) then
    -- Display a message to say there are no matches
    AchievementFrameAchievementsFeatOfStrengthText:SetText(L["No Matches"]);
    AchievementFrameAchievementsFeatOfStrengthText:Show();
  else
    AchievementFrameAchievementsFeatOfStrengthText:Hide();
  end

  -- If there is a selection...
  if (selection) then
    -- ...reset the objectives
    AchievementButton_ResetObjectives();
  end

  -- Work out the scroll frame's extra height
  extraHeight =
    (scrollFrame.largeButtonHeight or ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT);

  -- Update the scroll frame's buttons
  numAchievements = filteredAchievements[0];
  for buttonIdx = 1, numButtons do
    button = scrollFrame.buttons[buttonIdx];
    index = buttonIdx+offset;
    if (index <= numAchievements) then
      AchievementButton_DisplayAchievement(button, catId, index, selection);
      displayedHeight = displayedHeight+button:GetHeight();
    else
      button:Hide();
    end
  end

  -- Update the scroll frame
  totalHeight = (numAchievements*ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT)+
    (extraHeight-ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT);
  HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);

  -- Check if there is a selection
  if (selection) then
    -- Update the selected achievement
    AchievementFrameAchievements.selection = selection;
  else
    -- Collapse any expanded button in the scroll frame
    HybridScrollFrame_CollapseButton(scrollFrame);
  end

  return;
end
