-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

--
-- Function to return the index for any undefined indexes
--
local function defaultValueIndex(table, index)
  return index;
end

-- Set the function for any undefined indexes
setmetatable(L, {__index = defaultValueIndex});

local NUM_INVITES_PER_PAGE = 10;
local TIMER_DELAY_MAX      = 3600;
local TIMER_DELAY_MIN      = 0.1;

local cityZones =
{
  [L["Dalaran"]]               = true,
  [L["Darnassus"]]             = true,
  [L["Ironforge"]]             = true,
  [L["City of Ironforge"]]     = true,
  [L["Orgrimmar"]]             = true,
  [L["Shattrath"]]             = true,
  [L["Shattrath City"]]        = true,
  [L["Shrine of Seven Stars"]] = true,
  [L["Shrine of Two Moons"]]   = true,
  [L["Silvermoon City"]]       = true,
  [L["Stormwind"]]             = true,
  [L["Stormwind City"]]        = true,
  [L["The Exodar"]]            = true,
  [L["Thunder Bluff"]]         = true,
  [L["Undercity"]]             = true
};

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to disable a Niggles character list frame and its descendants
--
local function characterListEnable(self)
  self:SetEnabled(true);
  return;
end

--
-- Function to enable a Niggles character list frame and its descendants
--
local function characterListDisable(self)
  self:SetEnabled(false);
  return;
end

--
-- Function to set the enabled state of a Niggles character list frame and
-- its descendants
--
local function characterListSetEnabled(self, enable)
  -- Local Variables
  local buttons;

  -- Save the enabled status
  self.isEnabled = enable;

  -- Enable/Disable the label
  self.label:SetFontObject(enable and self.label.normalFont or
    self.label.disableFont);

  -- Enable/Disable the scroll frame and its buttons
  self.scrollFrame.ScrollBar:SetEnabled(enable);
  buttons = self.scrollFrame.buttons;
  if (buttons ~= nil) then
    for idx = 1, #buttons do
      buttons[idx]:SetEnabled(enable);
    end
  end
  self.scrollFrame.update();

  -- Enable/Disable the edit box and Add button
  self.edit:SetEnabled(enable);
  self.add:SetEnabled(enable);

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

NigglesHolidayBrewfest        = 0x00000001;
NigglesHolidayChildrensWeek   = 0x00000002;
NigglesHolidayDarkmoonFaire   = 0x00000004;
NigglesHolidayDayOfTheDead    = 0x00000008;
NigglesHolidayHallowsEnd      = 0x00000010;
NigglesHolidayHarvestFestival = 0x00000020;
NigglesHolidayLoveInTheAir    = 0x00000040;
NigglesHolidayLunarFestival   = 0x00000080;
NigglesHolidayMidsummer       = 0x00000100;
NigglesHolidayNobleGarden     = 0x00000200;
NigglesHolidayPiratesDay      = 0x00000400;
NigglesHolidayWinterVeil      = 0x00000800;

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to initialise a Niggles Character List
--
function NigglesCharacterListInit(frame, label, buttonTemplate, updateFunc, addFunc, dropDownFunc)
  -- Save some of the arguments
  frame.buttonTemplate = buttonTemplate;
  frame.addFunc        = addFunc;

  -- Set the frame's label
  frame.label:SetText(label);
  frame.label.normalFont  = "GameFontNormal";
  frame.label.disableFont = "GameFontDisable";

  -- Initialise the scroll frame
  frame.scrollFrame.update = updateFunc;
  frame.scrollFrame.scrollBar.doNotHide = true;

  -- Adjust some colours
  frame.scrollFrame.ScrollBar.trackBG:SetVertexColor(0.2, 0.13, 0.08, 0.25);
  frame.border:SetBackdropBorderColor(0.75, 0.75, 0.75, 1.0);

  -- Initialise the drop down menu
  if (dropDownFunc ~= nil) then
    frame.dropDown.info = UIDropDownMenu_CreateInfo();
    UIDropDownMenu_Initialize(frame.dropDown, dropDownFunc, "MENU");
  end

  return;
end

--
-- Function to process the Enter key in an Character List edit box
--
function NigglesCharacterListEditOnEnterPressed(self)
  -- Local Variables
  local name;
  local frame = self:GetParent();

  -- Check the AutoComplete frame doesn't want the key press
  if (not AutoCompleteEditBox_OnEnterPressed(self)) then
    -- Get the name entered by the user
    name = strtrim(self:GetText());
    if (name ~= "") then
      -- Capitalize the name
      name = (name:sub(1,1):upper())..(name:sub(2):lower());

      -- Call the add function
      if (frame.addFunc ~= nil) then
        frame.addFunc(frame, name);
      end

      -- Clear the edit frame
      self:SetText("");
      AutoComplete_HideIfAttachedTo(self)
    else
      self:ClearFocus();
    end
  end

  return;
end

--
-- Function to tidy up when a Character List Edit frame loses focus
--
function NigglesCharacterListEditOnFocusLost(self)
  AutoCompleteEditBox_OnEditFocusLost(self);
  self:HighlightText(0, 0);
  local trimmedText = strtrim(self:GetText());
  if (trimmedText == "") then
    self:SetText("");
  end

  return;
end

--
-- Function called when a Niggles Character List scroll frame changes size
--
-- NOTE: The buttons are created when the frame changes size so the character
--       list doesn't have to be fixed
--
function NigglesCharacterListScrollFrameOnSize(self)
  -- Local Variables
  local parent = self:GetParent();
  local isEnabled;
  local buttons;

  -- Check a button template has been supplied
  if (parent.buttonTemplate ~= nil) then
    -- Create buttons, if required
    HybridScrollFrame_CreateButtons(self,
      self:GetParent().buttonTemplate, 1, 0);

    -- Enable/Disable the buttons
    buttons = self.buttons;
    for idx = 1, #buttons do
      buttons[idx]:SetEnabled(parent.isEnabled);
    end

    -- Update the scroll frame, if possible
    if (self.update ~= nil) then
      self.update();
    end
  end

  return;
end

--
-- Function called when a Niggles Character List is created
--
function NigglesCharacterListOnLoad(self)
  -- Add some extra methods to the frame
  self.Disable    = characterListDisable;
  self.Enable     = characterListEnable;
  self.SetEnabled = characterListSetEnabled;

  return;
end

--
-- Function to update the names listed in an Calendar Events Invites list
--
function NigglesCharacterListUpdate(scrollbar)
  -- Local Variables
  local button;
  local frameName;
  local frame = scrollbar:GetParent();
  local numNames = 0;
  local topLine;

  -- Check the required variables are initialised
  if ((frame ~= nil) and (frame.list ~= nil)) then
    -- Get the frame's name
    frameName = frame:GetName();

    -- Get the number of names in the list
    numNames = #(frame.list);

    -- Make sure the top line displayed is still valid
    topLine = FauxScrollFrame_GetOffset(scrollbar);
    if (topLine+NUM_INVITES_PER_PAGE > numNames) then
      -- Work out the new top line
      topLine = numNames-NUM_INVITES_PER_PAGE;
      if (topLine < 0) then
        topLine = 0;
      end

      -- Set the new top line
      FauxScrollFrame_SetOffset(scrollbar, topLine);
    end

    -- Update all the buttons
    for buttonIdx = 1, NUM_INVITES_PER_PAGE do
      -- Get the current button
      button = _G[frameName..buttonIdx];

      -- Work out the index of the invite to display in the button
      if (topLine+buttonIdx <= numNames) then
        button:SetText(frame.list[topLine+buttonIdx]);
        button:Show();
        button:Enable();
      else
        button:Hide();
        button:Disable();
      end
    end

    -- Update the scrollbar
    FauxScrollFrame_Update(scrollbar, numNames, NUM_INVITES_PER_PAGE, 14,
      nil, 0, 0, nil, 0, 0, true);
  end

  return;
end

--
-- Function to enable/disable a frame and all its descendants. This function
-- is recursive.
--
function NigglesEnableFrame(frame, enable)
  -- Local Variables
  local childType;
  local frameType;

  -- Enable/disable the frame, if possible
  if ((frame.Enable) and (enable == true)) then
    frame:Enable();
  elseif ((frame.Disable) and (enable == false)) then
    frame:Disable();
  elseif (frame.GetObjectType) then
    frameType = frame:GetObjectType();
    if (frameType == "Frame") then
      -- Process each region
      for _, child in ipairs({frame:GetRegions()}) do
        -- Get the child's type
        childType = child:GetObjectType();

        -- Process the child, based on its type
        if (childType == "FontString") then
          if ((child.normalFont ~= nil) and (enable == true)) then
            child:SetFontObject(child.normalFont);
          elseif ((child.disableFont ~= nil) and (enable == false)) then
            child:SetFontObject(child.disableFont);
          end
        end
      end
    elseif (frameType == "EditBox") then
      frame:EnableJoystick(enable);
      frame:EnableMouse(enable);
      frame:EnableMouseWheel(enable);
      frame:EnableKeyboard(enable);
      frame:ClearFocus();
      if ((frame.normalFont ~= nil) and (enable == true)) then
        frame:SetFontObject(frame.normalFont);
      elseif ((frame.disableFont ~= nil) and (enable == false)) then
        frame:SetFontObject(frame.disableFont);
      end
    end
  end

  -- Process all children of the current frame
  for _, child in ipairs({frame:GetChildren()}) do
    NigglesEnableFrame(child, enable);
  end

  return;
end

--
-- Function to enable or disable a list of frames
--
function NigglesEnableFrames(enable, ...)
  -- Enable or disable the list of frames
  if (enable) then
    for _, frame in ipairs({...}) do
      frame:Enable();
    end
  else
    for _, frame in ipairs({...}) do
      frame:Disable();
    end
  end

  return;
end

--
-- Function to get the localized version of a string
--
function NigglesGetLocaleString(string)
  return L[string];
end

--
-- Function to check if a specified zone is a city
--
function NigglesIsCity(zoneName)
  return (cityZones[zoneName] ~= nil);
end

--
-- Function called each time the Niggles frame is redrawn to the screen. It is
-- used to do periodic tasks.
--
function NigglesOnUpdate(self, elapsed)
  -- Local Variables
  local currentTime = GetTime();

  -- Check if the next timer has expired
  if (Niggles.timers.nextExpire <= currentTime) then
    -- Local Variables
    local currentTimer;

    -- Reset the time when the next timer expires
    Niggles.timers.nextExpire = currentTime+TIMER_DELAY_MAX+1;

    -- Process the Niggles.timers
    currentTimer = Niggles.timers.list;
    while (currentTimer ~= nil) do
      -- Check if the timer has expired
      if (currentTimer.time <= currentTime) then
        -- Check the timer hasn't been marked for deletion
        if (not currentTimer.isDeleted) then
          -- Check if the timer is a repeating timer
          if (currentTimer.repeating == true) then
            -- Reset the timer
            currentTimer.time = currentTime+currentTimer.delay;
          else
            -- Mark the timer for deletion
            currentTimer.isDeleted = true;
          end

          -- Call the timer's function
          currentTimer.func(currentTimer.arg1, currentTimer.arg2);
        end

        -- Check if the timer is marked for deletion
        if (currentTimer.isDeleted) then
          -- Remove the timer from the list
          if (currentTimer.prev ~= nil) then
            currentTimer.prev.next = currentTimer.next;
          else
            Niggles.timers.list = currentTimer.next;
          end
          if (currentTimer.next ~= nil) then
            currentTimer.next.prev = currentTimer.prev;
          end
        end
      end

      -- Update the time of the next timer to expire
      if (not currentTimer.isDeleted) then
        Niggles.timers.nextExpire =
          math.min(Niggles.timers.nextExpire, currentTimer.time);
      end

      -- Move onto the next timer
      currentTimer = currentTimer.next;
    end

    -- Stop the monitoring event, if no longer required
    if ((Niggles.timers.nextExpire == currentTime+TIMER_DELAY_MAX+1) or
        (Niggles.timers.list == nil)) then
      Niggles:SetScript("OnUpdate", nil);
      Niggles.timers.nextExpire = 0;
    end
  end

  return;
end

--
-- Function to get all the points set for a frame and store them in an array
--
function NigglesFramePointsGet(frame, array)
  -- Clear the array
  wipe(array);

  -- Get all the frame's points
  for pointIdx = 1, frame:GetNumPoints() do
    array[pointIdx] = {frame:GetPoint(pointIdx)};
  end

  return;
end

--
-- Function to set all the points specified in an array for a frame
--
function NigglesFramePointsSet(frame, array)
  -- Clear all the frame's points
  frame:ClearAllPoints();

  -- Set all the frame's points
  for pointIdx = 1, #array do
    frame:SetPoint(array[pointIdx][1], array[pointIdx][2], array[pointIdx][3],
      array[pointIdx][4], array[pointIdx][5]);
  end

  return;
end

--
-- Function to check if a value is in a table, ignoring case
--
function NigglesTableContains(table, value)
  -- Local Variables
  local idx = 1;

  -- Search the table for the specified value
  value = value:lower();
  while (table[idx]) do
    if (value == table[idx]:lower()) then
      return 1;
    end
    idx = idx+1;
  end

  return nil;
end

--
-- Function to create a timer
--
function NigglesTimerCreate(name, func, arg1, arg2, delay, repeating)
  -- Local Variables
  local currentTimer;

  -- Make sure the timer's delay is valid
  delay = math.max(math.min(delay, TIMER_DELAY_MAX), TIMER_DELAY_MIN);

  -- Search for the named timer
  currentTimer = Niggles.timers.list;
  while (currentTimer ~= nil) do
    -- Check if the timer is the one specified
    if (currentTimer.name == name) then
      break;
    end
    currentTimer = currentTimer.next;
  end

  -- Check no existing timer was found
  if (currentTimer == nil) then
    -- Create a table for the new timer
    currentTimer =
    {
      name = name
    };

    -- Add the timer to the linked list
    if (Niggles.timers.list == nil) then
      Niggles.timers.list = currentTimer;
    else
      Niggles.timers.list.prev = currentTimer;
      currentTimer.next        = Niggles.timers.list;
      Niggles.timers.list      = currentTimer;
    end
  end

  -- (Re)initialise the existing timer
  currentTimer.func      = func;
  currentTimer.arg1      = arg1;
  currentTimer.arg2      = arg2;
  currentTimer.delay     = delay;
  currentTimer.repeating = repeating;
  currentTimer.time      = GetTime()+delay;
  currentTimer.isDeleted = false;

  -- Update when the next timer expires
  Niggles.timers.nextExpire =
    math.min(Niggles.timers.nextExpire, currentTimer.time);

  -- Start the monitoring event, if required
  if (Niggles:GetScript("OnUpdate") == nil) then
    Niggles:SetScript("OnUpdate", NigglesOnUpdate);
  end

  return;
end

--
-- Function to destroy a timer.
--
-- NOTE: The timer is actually destroyed in the monitoring event function to
--       allow timers to be safely destroyed within their own callback function.
--
function NigglesTimerDestroy(name)
  -- Local Variables
  local currentTimer = Niggles.timers.list;

  -- Search for the named timer
  while (currentTimer ~= nil) do
    -- Check if the timer is the one specified
    if (currentTimer.name == name) then
      -- Expire the timer and mark it for deletion
      currentTimer.time      = 0;
      currentTimer.delay     = 0;
      currentTimer.repeating = false;
      currentTimer.isDeleted = true;
      break;
    end

    -- Move onto the next timer
    currentTimer = currentTimer.next;
  end

  return;
end
